/*
 * x86work library for Win32
 * 05.07.24 : Makoto NARA (Mc.N)
 */

/*
 * Copyright (c) 2005 Makoto NARA (Mc.N) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY Makoto NARA (Mc.N) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "x86work.h"

#ifndef _NTDEF_
typedef LONG NTSTATUS, *PNTSTATUS;
#define STATUS_SUCCESS	((NTSTATUS)0x00000000L)
#endif

// MSR for Win32 struct
typedef struct _MSR_WIN32{
	DWORD num;
	DWORD reserved;
	DWORD lo;
	DWORD hi;
} MSR_WIN32, *PMSR_WIN32;

typedef enum _DEBUG_CONTROL_CODE {
	DebugSysReadIoSpace = 14,
	DebugSysWriteIoSpace = 15,
	DebugSysReadMsr = 16,
	DebugSysWriteMsr = 17,
	DebugSysReadBusData = 18,
	DebugSysWriteBusData = 19,
} DEBUG_CONTROL_CODE;

// ZwSystemDebugControl
typedef NTSTATUS (NTAPI *PZWSYSTEMDEBUGCONTROL)(
	DEBUG_CONTROL_CODE ControlCode,
    PVOID InputBuffer,
	ULONG InputBufferLength,
	PVOID OutputBuffer,
    ULONG OutputBufferLength,
	PULONG ReturnLength
	);

// OpenProcessToken()
typedef BOOL (*POPENPROCESSTOKEN)(
	HANDLE ProcessHandle,
	DWORD DesiredAccess,
	PHANDLE TokenHandle
	);

// AdjustTokenPrivileges
typedef BOOL (*PADJUSTTOKENPRIVILEGES)(
    HANDLE TokenHandle,
    BOOL DisableAllPrivileges,
    PTOKEN_PRIVILEGES NewState,
    DWORD BufferLength,
    PTOKEN_PRIVILEGES PreviousState,
    PDWORD ReturnLength
    );

// LookupPrivilegeValueA
typedef BOOL (*PLOOKUPPRIVILEGEVALUEA)(
    LPCSTR lpSystemName,
    LPCSTR lpName,
    PLUID   lpLuid
	);

BOOL
SetPrivilege(
    HANDLE hToken,
    LPCSTR szPrivilege,
    BOOL bEnablePrivilege
    );

PCPUIDS_DATA
EnumCPUIDs(
	DWORD numcpu,
	DWORD cntcpu
	);

PCPUIDS_DATA
EnumCPUID(
	DWORD numcpu,
	DWORD cntcpu,
	DWORD idx
	);

BOOL IsWin9x(void);

PCPUID_INFOR
GetCPUIDdataXP(
	void
	);

PCPUID_INFOR
GetCPUIDdata9x(
	void
	);





BOOL
_cpuid_os(
	DWORD idx,
	PCPUID_DATA pcpu
	)
{
	return _cpuids_os(0,0,idx,pcpu);
}


BOOL
_cpuids_os(
	DWORD numcpu,
	DWORD cntcpu,
	DWORD idx,
	PCPUID_DATA pcpu
	)
{
	DWORD a, b, c, d;
	BOOL bRet = TRUE;

	if(pcpu==NULL){return FALSE;}
	// input
	a = idx;
	b = pcpu->ebx;
	c = pcpu->ecx;
	d = pcpu->edx;

	// init
	ZeroMemory(pcpu, sizeof(CPUID_DATA));

	__try{
	_asm{
		pushad
		mov eax, a
		mov ebx, b
		mov ecx, c
		mov edx, d
		cpuid
		mov a, eax
		mov b, ebx
		mov c, ecx
		mov d, edx
		popad
	}
	}__except(EXCEPTION_EXECUTE_HANDLER){
		bRet = FALSE;
	}
	if( bRet ){
		pcpu->numcpu = numcpu;
		pcpu->cntcpu = cntcpu;
		pcpu->idx = idx;
		pcpu->eax = a;
		pcpu->ebx = b;
		pcpu->ecx = c;
		pcpu->edx = d;
	}

	return bRet;
}




BOOL
_rdmsr_os(
	DWORD idx,
	PMSR_DATA pmsr
	)
{
	BOOL bRet = FALSE;
	HMODULE hNTDLL = NULL, hADVAPI = NULL;
	PZWSYSTEMDEBUGCONTROL pfnZwSystemDebugControl = NULL;
	POPENPROCESSTOKEN     pfnOpenProcessToken = NULL;
	HANDLE hToken = NULL;
	MSR_WIN32 msrw32;
	NTSTATUS ntstat;

	// init
	if( pmsr == NULL ){ return FALSE; }
	ZeroMemory(pmsr, sizeof(MSR_DATA));
	ZeroMemory(&msrw32, sizeof(MSR_WIN32));
	bRet = FALSE;

	hNTDLL  = LoadLibrary(TEXT("ntdll.dll"));
	hADVAPI = LoadLibrary(TEXT("advapi32.dll"));
	if( (hNTDLL==NULL)||(hADVAPI==NULL) ){ goto cleanup; }

	pfnZwSystemDebugControl = (PZWSYSTEMDEBUGCONTROL)GetProcAddress(hNTDLL, TEXT("ZwSystemDebugControl"));
	pfnOpenProcessToken     = (POPENPROCESSTOKEN)GetProcAddress(hADVAPI, TEXT("OpenProcessToken"));
	if( NULL == pfnZwSystemDebugControl || NULL == pfnOpenProcessToken ){ goto cleanup; }

	// open Token
	if( !pfnOpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken) ){ goto cleanup; }

	// enable SeDebugPrivilege
	if( !SetPrivilege(hToken, SE_DEBUG_NAME, TRUE) ){ goto cleanup;	}

	// rdmsr
	msrw32.num = idx;
	ntstat = pfnZwSystemDebugControl( DebugSysReadMsr, &msrw32, sizeof(msrw32), NULL, 0, NULL );
	if( STATUS_SUCCESS == ntstat ){
		pmsr->edx = msrw32.hi;
		pmsr->eax = msrw32.lo;
		bRet = TRUE;
	}

	// disable SeDebugPrivilege
	SetPrivilege(hToken, SE_DEBUG_NAME, FALSE);

cleanup:
	if(NULL!=hToken){CloseHandle(hToken);}
	if(NULL!=hADVAPI){FreeLibrary(hADVAPI);}
	if(NULL!=hNTDLL){FreeLibrary(hNTDLL);}

	return bRet;
}




// [HOWTO] SeDebugPrivilege gpĔCӂ̃vZXւ̃nh擾@
// MSKB : 131065
// http://support.microsoft.com/default.aspx?scid=kb;ja;131065
BOOL
SetPrivilege(
    HANDLE hToken,
    LPCSTR szPrivilege,
    BOOL bEnablePrivilege
    )
{
    TOKEN_PRIVILEGES tp;
    LUID luid;
    TOKEN_PRIVILEGES tpPrevious;
    DWORD cbPrevious=sizeof(TOKEN_PRIVILEGES);
	BOOL bRet = FALSE;
	HMODULE hNTDLL = NULL, hADVAPI = NULL;
	PLOOKUPPRIVILEGEVALUEA pfnLookupPrivilegeValueA = NULL;	// LookupPrivilegeValue  (advapi32.lib)
	PADJUSTTOKENPRIVILEGES pfnAdjustTokenPrivileges = NULL; // AdjustTokenPrivileges (advapi32.lib)

	hADVAPI = LoadLibrary(TEXT("advapi32.dll"));
	if( hADVAPI == NULL ){ bRet = FALSE; goto cleanup; }
	pfnLookupPrivilegeValueA = (PLOOKUPPRIVILEGEVALUEA)GetProcAddress(hNTDLL, TEXT("LookupPrivilegeValue"));
	pfnAdjustTokenPrivileges = (PADJUSTTOKENPRIVILEGES)GetProcAddress(hNTDLL, TEXT("AdjustTokenPrivileges"));
	if( NULL==pfnLookupPrivilegeValueA || NULL==pfnAdjustTokenPrivileges ){ bRet = FALSE; goto cleanup; }

    if( !pfnLookupPrivilegeValueA( NULL, szPrivilege, &luid ) ){ bRet = FALSE; goto cleanup; }

    // get current privilege setting
	ZeroMemory(&tp, sizeof(tp));
    tp.PrivilegeCount           = 1;
    tp.Privileges[0].Luid       = luid;
    tp.Privileges[0].Attributes = 0;

    bRet = pfnAdjustTokenPrivileges(
            hToken,
            FALSE,
            &tp,
            sizeof(TOKEN_PRIVILEGES),
            &tpPrevious,
            &cbPrevious
            );

	if( !bRet ){ goto cleanup; }

	// set privilege based on previous setting
    tpPrevious.PrivilegeCount       = 1;
    tpPrevious.Privileges[0].Luid   = luid;

    if(bEnablePrivilege) {
        tpPrevious.Privileges[0].Attributes |= (SE_PRIVILEGE_ENABLED);
    }else{
        tpPrevious.Privileges[0].Attributes ^= (SE_PRIVILEGE_ENABLED & tpPrevious.Privileges[0].Attributes);
    }

    bRet = pfnAdjustTokenPrivileges(
            hToken,
            FALSE,
            &tpPrevious,
            cbPrevious,
            NULL,
            NULL
            );

cleanup:
	if(NULL!=hADVAPI){FreeLibrary(hADVAPI);}

    return bRet;
}




BOOL
IsSupportedNXOS(
	void
	)
{
	OSVERSIONINFOEX osvie;
	ZeroMemory(&osvie, sizeof(OSVERSIONINFOEX));

	osvie.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	if( !GetVersionEx((POSVERSIONINFO)&osvie) ){
		return FALSE;
	}

	// NT base?
	if( osvie.dwPlatformId != VER_PLATFORM_WIN32_NT ){ return FALSE; }
	// above xp sp2? ---> SUPPORT!
	if( (osvie.dwMajorVersion==5) && (osvie.dwMinorVersion==1) && (osvie.wServicePackMajor>=2) ){
		return TRUE;
	}
	// above w2k3 sp1? ---> SUPPORT!
	if( (osvie.dwMajorVersion==5) && (osvie.dwMinorVersion==2) && (osvie.wServicePackMajor>=1) ){
		return TRUE;
	}
	//  Major Version is above 6? ---> SUPPORT, maybe...
	if( osvie.dwMajorVersion>5 ){
		return TRUE;
	}
	// Major=5, Minor>2  ---> SUPPORT, maybe...
	if( (osvie.dwMajorVersion==5) && (osvie.dwMinorVersion>2) ){
		return TRUE;
	}

	// w2k or xp sp1
	return FALSE;
}




typedef int (*PFUNC)(void);
BOOL
IsFindOV(
	void
	)
{
	BOOL bRet = FALSE;
	PFUNC pfunc = NULL;
	PBYTE pf = NULL;
	DWORD dwProt = 0;

	// Read/Write/Execute
	pf = (PBYTE)VirtualAlloc( NULL, 4, MEM_COMMIT, PAGE_EXECUTE_READWRITE );
	if( pf == NULL ){ return FALSE; }
	*pf     = 0xC3;	// ret*4
	*(pf+1) = 0xC3;
	*(pf+2) = 0xC3;
	*(pf+3) = 0xC3;

	pfunc = (PFUNC)pf;

	// SEH
	__try{
		(*pfunc)();
	}__except(EXCEPTION_EXECUTE_HANDLER){
		// unknown...
		goto cleanup;
	}

	// Read/Write, NO Execute
	if( VirtualProtect(pf, 4, PAGE_READWRITE, &dwProt) ){
		__try{
			(*pfunc)();
		}__except(EXCEPTION_EXECUTE_HANDLER){
			// work NX-bit!
			bRet = TRUE;
		}
	}

cleanup:
	if( pf != NULL ){ VirtualFree( pf, 0, MEM_DECOMMIT | MEM_RELEASE ); }
	return bRet;
}


PCPUIDS_DATA
EnumCPUIDs(
	DWORD numcpu,
	DWORD cntcpu
	)
{
	PCPUIDS_DATA p, ptmp, porg =  NULL;
	DWORD dw, dwValue = 0;
	BOOL bRet = FALSE;

	// get CPUID(0)
	if( NULL == (p = EnumCPUID(numcpu, cntcpu, 0)) ){ goto cleanup; }
	porg = p;

	dwValue = p->cpu.eax;
	for( dw=1; dw<(dwValue+1); dw++ ){
		if( NULL == (ptmp = EnumCPUID(numcpu, cntcpu, dw)) ){ goto cleanup; }
		p->pnext = ptmp;
		p = ptmp;
	}

	// get extended CPUID
	if( NULL == (ptmp = EnumCPUID(numcpu, cntcpu, EXTCPUID+0)) ){ goto cleanup; }
	dwValue = ptmp->cpu.eax;
	if( !(dwValue>=0x80000000) ){
		// not supported extended CPUID
		CloseCPUIDdata(ptmp);
		bRet = TRUE;
		goto cleanup;
	}
	// connect!
	p->pnext = ptmp;
	p = ptmp;
	for( dw=1; dw<(dwValue-EXTCPUID+1); dw++ ){
		if( NULL == (ptmp = EnumCPUID(numcpu, cntcpu, EXTCPUID+dw)) ){ goto cleanup; }
		p->pnext = ptmp;
		p = ptmp;
	}
	bRet = TRUE;

cleanup:
	if(!bRet){
		if(NULL!=porg){ CloseCPUIDdata(porg); }
		porg = NULL;
	}

	return porg;
}


PCPUIDS_DATA
EnumCPUID(
	DWORD numcpu,
	DWORD cntcpu,
	DWORD idx
	)
{
	PCPUIDS_DATA p =  NULL;
	if( NULL == (p = (PCPUIDS_DATA)malloc(sizeof(CPUIDS_DATA))) ){ return NULL; }
	memset(p,0,sizeof(CPUIDS_DATA));
	
	if( !_cpuids_os(numcpu,cntcpu,idx,&(p->cpu)) ){
		free(p);
		p = NULL;
	}
	return p;
}


BOOL
IsWin9x(void)
{
	OSVERSIONINFOEX osvie;
	ZeroMemory(&osvie, sizeof(OSVERSIONINFOEX));

	osvie.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	if( !GetVersionEx((POSVERSIONINFO)&osvie) ){
		return FALSE;
	}
	if( VER_PLATFORM_WIN32_WINDOWS == osvie.dwPlatformId ){
		return TRUE;
	}
	return FALSE;
}

PCPUID_INFOR
GetCPUIDdataXP(
	void
	)
{
	HANDLE hCP;
	DWORD pamask = 0, samask = 0, patmp = 0;
	int cntcpu = 0, numcpu = 0;
	BOOL bRet = FALSE;
	PCPUID_INFOR pci = NULL;
	PCPUIDS_DATA p, ptmp;

	if( NULL == (pci = (PCPUID_INFOR)malloc(sizeof(CPUID_INFOR))) ){ goto cleanup; }
	memset(pci, '\0', sizeof(CPUID_INFOR) );

	// How many processors does it have?
	hCP = GetCurrentProcess();
	if( !GetProcessAffinityMask( hCP, &pamask, &samask ) ){ goto cleanup; }
	if( 0==samask || 0==pamask ){ goto cleanup; }

	numcpu = 0; cntcpu = 0;
	bRet = FALSE;
	for(;;){
		patmp = samask>>numcpu;
		// no more cpu?
		if( 0 == patmp ){
			bRet = TRUE;
			break;
		}
		// 
		if( !(patmp&0x1) ){
			numcpu++;
			continue;
		}

		// set process#
		if( !SetProcessAffinityMask(hCP, 1<<numcpu) ){
			break;
		}
		Sleep(1);
		
		// get CPUIDs
		if( NULL== (p = EnumCPUIDs(numcpu, cntcpu)) ){ break; }
		if( NULL == pci->pcpus ){
			pci->pcpus = p;
		}else{
			ptmp = pci->pcpus;
			for(;;){
				if( ptmp->pnext == NULL ){
					ptmp->pnext = p;
					break;
				}
				ptmp = ptmp->pnext;
			}
		}
		// count cpu!
		cntcpu++;
		numcpu++;
	}
	if( bRet ){
		pci->dwProcessor = cntcpu;
	}

cleanup:
	if(!bRet){
		if( NULL != pci ){
			CloseCPUIDinfo(pci);
		}
		pci = NULL;
	}

	return pci;
}

PCPUID_INFOR
GetCPUIDdata9x(
	void
	)
{
	BOOL bRet = FALSE;
	PCPUID_INFOR pci = NULL;

	if( NULL == (pci = (PCPUID_INFOR)malloc(sizeof(CPUID_INFOR))) ){ goto cleanup; }
	memset(pci, '\0', sizeof(CPUID_INFOR) );

	// get CPUIDs for Win9x/Me
	pci->pcpus = EnumCPUIDs(0,0);

	if( NULL == pci->pcpus ){
		CloseCPUIDinfo(pci);
		pci = NULL;
	}else{
		pci->dwProcessor = 1;
	}

cleanup:
	return pci;
}

PCPUID_INFOR
GetCPUIDdata_os(
	void
	)
{
	PCPUID_INFOR pci = NULL;

	if( IsWin9x() ){
		pci = GetCPUIDdata9x();
	}else{
		pci = GetCPUIDdataXP();
	}
	
	return pci;
}

